/*
 * (c) Copyright 2010-2011 by Volker Bergmann. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted under the terms of the
 * GNU General Public License (GPL).
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * WITHOUT A WARRANTY OF ANY KIND. ALL EXPRESS OR IMPLIED CONDITIONS,
 * REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE
 * HEREBY EXCLUDED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

package org.databene.benerator.engine.statement;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.databene.benerator.engine.BeneratorContext;
import org.databene.commons.CollectionUtil;
import org.databene.commons.ConfigurationError;
import org.databene.commons.ErrorHandler;
import org.databene.commons.Expression;
import org.databene.commons.collection.OrderedNameMap;
import org.databene.commons.expression.ExpressionUtil;
import org.databene.dbsanity.parser.TestSuiteParser;
import org.databene.jdbacl.identity.IdentityModel;
import org.databene.jdbacl.identity.IdentityProvider;
import org.databene.jdbacl.identity.KeyMapper;
import org.databene.jdbacl.identity.NoIdentity;
import org.databene.jdbacl.identity.mem.MemKeyMapper;
import org.databene.jdbacl.model.DBTable;
import org.databene.jdbacl.model.Database;
import org.databene.model.data.ComplexTypeDescriptor;
import org.databene.model.data.ReferenceDescriptor;
import org.databene.platform.db.DBSystem;

/**
 * Groups {@link TranscodeStatement}s and provides common features like 
 * {@link IdentityProvider} and {@link KeyMapper} objects.<br/><br/>
 * Created: 10.09.2010 18:25:18
 * @since 0.6.4
 * @author Volker Bergmann
 */
public class TranscodingTaskStatement extends SequentialStatement {
	
    Expression<DBSystem> sourceEx; 
    Expression<DBSystem> targetEx;
    Expression<String> identityEx;
	Expression<Long> pageSizeEx;
    Expression<ErrorHandler> errorHandlerExpression;
    IdentityProvider identityProvider;
    KeyMapper mapper;
	Map<String, Boolean> tableNkRequirements = OrderedNameMap.createCaseIgnorantMap();
    
	public TranscodingTaskStatement(Expression<DBSystem> sourceEx, Expression<DBSystem> targetEx, Expression<String> identityEx, 
    		Expression<Long> pageSizeEx, Expression<ErrorHandler> errorHandlerExpression) {
	    this.sourceEx = cache(sourceEx);
	    this.targetEx = cache(targetEx);
	    this.identityEx = cache(identityEx);
	    this.pageSizeEx = cache(pageSizeEx);
	    this.errorHandlerExpression = cache(errorHandlerExpression);
		this.identityProvider = new IdentityProvider();
    }

	public Expression<DBSystem> getSourceEx() {
	    return sourceEx;
    }

	public Expression<DBSystem> getTargetEx() {
	    return targetEx;
    }

    public Expression<Long> getPageSizeEx() {
	    return pageSizeEx;
    }

	public Expression<ErrorHandler> getErrorHandlerEx() {
    	return errorHandlerExpression;
    }

	public IdentityProvider getIdentityProvider() {
		return identityProvider;
	}
	
	KeyMapper getKeyMapper() {
		return mapper;
	}
	
	@Override
	public void execute(BeneratorContext context) {
		DBSystem target = getTarget(context);
		mapper = new MemKeyMapper(null, null, target.getConnection(), target.getId(), identityProvider);
		checkPrecoditions(context);
		super.execute(context);
	}
	
	private void checkPrecoditions(BeneratorContext context) {
		boolean identitiesRequired = false;
		DBSystem target = targetEx.evaluate(context);
		List<TranscodeStatement> transcodes = CollectionUtil.extractItemsOfType(TranscodeStatement.class, subStatements);
		for (TranscodeStatement transcode : transcodes) {
			ComplexTypeDescriptor type = transcode.getType(context);
			String tableName = type.getName();
			// require that the target table is empty
			if (target.countEntities(tableName) > 0)
				throw new ConfigurationError("Error transcoding table '" + tableName + "' to " + target + ": target table is not empty");
			// items to be transcoded do not need NK definition
			tableNkRequirements.put(tableName, false);
			for (ReferenceDescriptor ref : type.getReferenceComponents()) {
				String targetTable = ref.getTargetType();
				if (!tableNkRequirements.containsKey(targetTable)) {
					tableNkRequirements.put(targetTable, true);
					identitiesRequired = true;
				}
			}
		}
		// check that each table for which an identity definition is required has one
		if (identitiesRequired)
			readIdentityDefinition(context);
		for (Entry<String, Boolean> req : tableNkRequirements.entrySet()) {
			String tableName = req.getKey();
			Boolean required = req.getValue();
			IdentityModel identity = identityProvider.getIdentity(tableName, false);
			if (identity == null) {
				if (required)
					throw new ConfigurationError("For transcoding, an identity definition of table " + tableName + " is required");
				else {
					DBTable table = target.getDbMetaData().getTable(tableName);
					identityProvider.registerIdentity(new NoIdentity(table), tableName);
				}
			}
		}
	}
	
	
	
	// helpers ---------------------------------------------------------------------------------------------------------

	private void readIdentityDefinition(BeneratorContext context) {
		try {
			// check identity definition
			String identityUri = ExpressionUtil.evaluate(identityEx, context);
			if (identityUri == null)
				throw new ConfigurationError("No 'identity' definition file defined");
			String idFile = context.resolveRelativeUri(identityUri);
			DBSystem target = getTarget(context);
			Database database = target.getDbMetaData();
			File reportFolder = new File("dbsanity-report");
			new TestSuiteParser().parseHierarchy(new File(idFile), reportFolder, reportFolder, new File("temp"), database, mapper, identityProvider);
		} catch (Exception e) {
			throw new ConfigurationError("Error setting up transcoding task", e);
		}
	}

	private DBSystem getTarget(BeneratorContext context) {
		DBSystem target = ExpressionUtil.evaluate(targetEx, context);
		if (target == null)
			throw new ConfigurationError("No 'target' database defined in <transcodingTask>");
		return target;
	}

	public boolean needsNkMapping(String tableName) {
		Boolean required = tableNkRequirements.get(tableName);
		if (required == null)
			throw new RuntimeException("Assertion failed: Not clear if an identity definition is necessary for table " + tableName);
		return required;
	}
	
}
